/*
 * bbornpv.cpp
 *
 *  Created on: Apr 28, 2017
 *      Author: joe
 */

#include <random>
#include "bbornpv.h"
#include "laplacian10thispc.h"
#include "zeropad.h"
#include "afd2dfst.h"
#include "sourcescale.h"
#include "bscat.h"
#include "scndder.h"
#include "pvrrec.h"
#include "blendnc2d.h"
#include "waveletpad.h"

bbornpv::bbornpv(std::vector<axis> paxes, std::vector<axis> daxes, axis saxis, int srcz, int recz,
    hypercube_float *bvel, hypercube_float *src, float rho, float dtu, float fmax,
    int minblnd, int maxblnd, float mindly, float maxdly, int seed, int space) {
  /* Get axes */
  _paxes = paxes; _daxes = daxes; _saxis = saxis;
  _nsx = _saxis.n; _osx = _saxis.o; _dsx = _saxis.d;
  _raxes.push_back(_daxes[0]); _raxes.push_back(_daxes[1]); _raxes.push_back(_daxes[2]);
  _bvel = bvel; _src = src;

  /* Get single parameters */
  _rho = rho; _dtu = dtu; _fmax = fmax;
  _recz = recz; _srcz = srcz;
  _nt = daxes[0].n; _nrx = daxes[1].n;
  _nz = _paxes[0].n; _nx = _paxes[1].n;
  _wfldj = 10;

  /* Check wavefield */
  if((int)_paxes.size() != 3) seperr("bornpv: needs a cube for 2D wave propagation.\n");
  /* Check data */
  if((int)_daxes.size() != 4) seperr("bornpv: data must have axes of nt, nr, nc and ns.\n");
  if(_daxes[2].n != 2) seperr("bornpv: assumes only two component data (particle velocities).\n");
  if(_daxes[3].n != _nsx) seperr("bornpv: data must have same number of shots as shot axis.\n");
  /* Check background velocity */
  if(!_bvel->get_axis(1).same_axis(_paxes[0]) || !_bvel->get_axis(2).same_axis(_paxes[1])) {
    seperr("bornpv: velocity axes must be same as spatial axes of wavefield.\n");
  }
  /* Check source */
  if(_src->get_ndim() != 1) seperr("bornpv: assumes the source is a 1D wavelet.\n");
  if(!_src->get_axis(1).same_axis(daxes[0])) seperr("bornpv: source axis must be same as data.\n");

  /* Create container of images */
  std::vector<axis> iaxes;
  iaxes.push_back(_paxes[0]);
  iaxes.push_back(_paxes[1]);
  iaxes.push_back(_daxes[3]);
  _imgs = new hypercube_float(iaxes,true);

  /* Build shot time array */
  std::random_device rd; std::mt19937 rng;
  if(!seed) {
    rng.seed(rd());
  } else {
    rng.seed(seed);
  }
  std::uniform_real_distribution<float> unif(maxdly,mindly);
  /* First time is always zero */
  _stimei.push_back(0.0);
  for(int is = 1; is < _nsx; ++is) {
    auto val = unif(rng);
    _stimei.push_back(val);
  }

  /* Build blending array */
  blendnc2d *gam = new blendnc2d(_nsx,minblnd,maxblnd);
  _nb = gam->get_nb();

  /* Create shot positions */
  int j = 0; int k = space;
  for(int is = 0; is < (int)_nb.size(); ++is) {
    std::vector<float> xsi;
    std::vector<float> stimej;
    if(is != 0 && is%space == 0) j += k;
    for(int ib = 0; ib < _nb[is]; ++ib) {
      xsi.push_back(_osx + _dsx*j);
      stimej.push_back(_stimei[j]);
      if(ib == 0) {
        if(j + k >= _nsx) k = 2; //FIXME: A hack for now
        j += k;
      }
    }
    j = j - k + 1;
    _xss.push_back(xsi);
    _stimes.push_back(stimej);
  }

//  for(int is = 0; is < (int)_nb.size(); ++is) {
//    std::vector<float> xsi = _xss[is];
//    std::vector<float> stimej = _stimes[is];
//    for(int ib = 0; ib < _nb[is]; ++ib) {
//      fprintf(stderr,"%f %f\n",xsi[ib],stimej[ib]);
//    }
//  }
}

bbornpv::bbornpv(std::vector<axis> paxes, std::vector<axis> daxes, int srcz, int recz,
    hypercube_float *bvel, hypercube_float *src, float rho, float dtu, float fmax,
    int minblnd, int maxblnd, std::vector<float> stimes, std::vector<float> xss) {
  /* Get axes */
  _paxes = paxes; _daxes = daxes;
  _nsx = xss.size(); 
  _raxes.push_back(_daxes[0]); _raxes.push_back(_daxes[1]); _raxes.push_back(_daxes[2]);
  _bvel = bvel; _src = src;

  /* Get single parameters */
  _rho = rho; _dtu = dtu; _fmax = fmax;
  _recz = recz; _srcz = srcz;
  _nt = daxes[0].n; _nrx = daxes[1].n;
  _nz = _paxes[0].n; _nx = _paxes[1].n;
  _wfldj = 10;

  /* Check wavefield */
  if((int)_paxes.size() != 3) seperr("bornpv: needs a cube for 2D wave propagation.\n");
  /* Check data */
  if((int)_daxes.size() != 4) seperr("bornpv: data must have axes of nt, nr, nc and ns.\n");
  if(_daxes[2].n != 2) seperr("bornpv: assumes only two component data (particle velocities).\n");
  /* Check background velocity */
  if(!_bvel->get_axis(1).same_axis(_paxes[0]) || !_bvel->get_axis(2).same_axis(_paxes[1])) {
    seperr("bornpv: velocity axes must be same as spatial axes of wavefield.\n");
  }
  /* Check source */
  if(_src->get_ndim() != 1) seperr("bornpv: assumes the source is a 1D wavelet.\n");
  if(!_src->get_axis(1).same_axis(daxes[0])) seperr("bornpv: source axis must be same as data.\n");

  /* Build blending array */
  blendnc2d *gam = new blendnc2d(_nsx,minblnd,maxblnd);
  _nb = gam->get_nb();
  if(_daxes[3].n != (int)(_nb.size())) seperr("bornpv: Calculated number of shots must be same as given.\n");

  /* Get shot times and positions */
  if(stimes.size() != xss.size()) {
    seperr("bbornpv: number of shot times %d should = number of shot positions %d\n",(int)stimes.size(),(int)xss.size());
  }
  _stimei = stimes; _xsi = xss;

  /* Create container of images */
  std::vector<axis> iaxes;
  iaxes.push_back(_paxes[0]);
  iaxes.push_back(_paxes[1]);
  iaxes.push_back(_daxes[3]);
  _imgs = new hypercube_float(iaxes,true);

  /* Create shot positions */
  int j = 0;
  for(int is = 0; is < (int)_nb.size(); ++is) {
    std::vector<float> xsi;
    std::vector<float> stimej;
    for(int ib = 0; ib < _nb[is]; ++ib) {
      xsi.push_back(_xsi[j]);
      stimej.push_back(_stimei[j]);
      ++j;
      }
    _xss.push_back(xsi);
    _stimes.push_back(stimej);
  }

//  for(int is = 0; is < (int)_nb.size(); ++is) {
//    std::vector<float> xsi = _xss[is];
//    std::vector<float> stimej = _stimes[is];
//    fprintf(stderr,"shotnum=%d\n",is);
//    for(int ib = 0; ib < _nb[is]; ++ib) {
//      fprintf(stderr,"%f %f\n",xsi[ib],stimej[ib]);
//    }
//  }
}

void bbornpv::setbwfldout(std::string wfname, int wfldj) {
  _bwfname = wfname; _wfldj = wfldj;
  _wfout = sregf(_paxes);
  _wfout.write_description(wfname);
}

void bbornpv::setawfldout(std::string wfname, int wfldj) {
  _awfname = wfname; _wfldj = wfldj;
  _awfout = sregf(_paxes);
  _awfout.write_description(wfname);
}

bool bbornpv::forward(bool add, my_vector *mod, my_vector *dat) {
  if(!add && dat) dat->zero();

  hypercube_float *m = (hypercube_float*)mod;
  hypercube_float *d = (hypercube_float*)dat;

  if(!m->check_hypercube_match(_bvel))  seperr("bornpv forward: input model perturbation must be same "
      "size as background velocity.\n");
  if(!d->get_axis(4).same_axis(_daxes[3])) seperr("bornpv forward: data axes must be same as constructor.\n");

  /* Loop over blended shots */
  int bw = 0;
#pragma omp parallel for default(shared)
  for(int is = 0; is < (int)_nb.size(); ++is) {
    laplacian10thispc *lp   = new laplacian10thispc(_paxes);
    sourcescale ss          = sourcescale(_dtu,_bvel);
    afd2dfst fd             = afd2dfst(_paxes,_bvel,lp,_dtu,_fmax);
    pvrrec pvrr             = pvrrec(_paxes,_raxes[1],_recz,_rho);
    scndder d2              = scndder(_raxes[0]);
    waveletpad wp           = waveletpad(_src->get_axis(1));
    hypercube_float *tpd = new hypercube_float(_src->get_axes(),true);
    hypercube_float *prc = new hypercube_float(_paxes,true);
    hypercube_float *bck = new hypercube_float(_paxes,true);
    hypercube_float *ssc = new hypercube_float(_paxes,true);
    hypercube_float *sht = new hypercube_float(_raxes,true);
    hypercube_float *rec = new hypercube_float(_raxes,true);
    bscat   im              = bscat(prc);
    /* Compute background wavefield */
    fprintf(stderr,"shotnum=%d\n",is);
    /* Build blended source */
    prc->zero();
    std::vector<float> xsi = _xss[is];
    std::vector<float> stimej = _stimes[is];
    for(int ib = 0; ib < _nb[is]; ++ib) {
      /* Pad each shot in time */
      wp.set_time(stimej[ib]);
      wp.forward(false,_src,tpd);
      fprintf(stderr,"xs=%f stime=%f\n",xsi[ib],stimej[ib]);
      zeropad zp = zeropad(static_cast<int>(xsi[ib]),_srcz);
      /* Blend the shots */
      zp.forward(true,tpd,prc);
    }
    ss.forward(false,prc,ssc);
    fd.forward(false,ssc,prc);
    /* If user specifies, write out background wavefield */
    im.forward(false,m,ssc);
    d2.forward(false,ssc,prc);
    ss.forward(false,prc,ssc);
    fd.forward(false,ssc,prc);
    /* If user specifies, write out born wavefield */
    if(!_bwfname.empty()) {
      if(is%_wfldj == 0) {
        _wfout.write_all(_bwfname,prc);
        _wfout.write_axis(_bwfname,4,axis(bw+1,0,1.0,"shot"));
        ++bw;
      }
    }
    pvrr.adjoint(false,rec,prc);
    memcpy(&d->vals[2*is*_nt*_nrx],rec->vals,sizeof(float)*2*_nt*_nrx);
    delete bck; delete ssc;
    delete sht; delete rec; delete tpd;
  }

  return true;
}

bool bbornpv::adjoint(bool add, my_vector *mod, my_vector *dat) {
  if(!add && mod) mod->zero();

  hypercube_float *m = (hypercube_float*)mod;
  hypercube_float *d = (hypercube_float*)dat;

  if(!m->check_hypercube_match(_bvel))  seperr("bornpv forward: input model perturbation must be same "
      "size as background velocity.\n");
  if(!d->get_axis(4).same_axis(_daxes[3])) seperr("bornpv forward: data axes must be same as constructor.\n");

  /* Loop over shot */
  int bw = 0, aw = 0;
#pragma omp parallel for default(shared)
  for(int is = 0; is < (int)_nb.size(); ++is) {
    laplacian10thispc *lp   = new laplacian10thispc(_paxes);
    sourcescale ss          = sourcescale(_dtu,_bvel);
    afd2dfst fd             = afd2dfst(_paxes,_bvel,lp,_dtu,_fmax);
    pvrrec pvrr             = pvrrec(_paxes,_raxes[1],_recz,_rho);
    scndder d2              = scndder(_raxes[0]);
    waveletpad wp           = waveletpad(_src->get_axis(1));
    hypercube_float *tpd = new hypercube_float(_src->get_axes(),true);
    hypercube_float *prc = new hypercube_float(_paxes,true);
    hypercube_float *bck = new hypercube_float(_paxes,true);
    hypercube_float *ssc = new hypercube_float(_paxes,true);
    hypercube_float *img = new hypercube_float(_bvel->get_axes(),true);
    hypercube_float *sht = new hypercube_float(_raxes,true);
    hypercube_float *rec = new hypercube_float(_raxes,true);
    bscat   im              = bscat(prc);
    /* Build blended source */
    prc->zero();
    std::vector<float> xsi = _xss[is];
    std::vector<float> stimej = _stimes[is];
    fprintf(stderr,"shotnum=%d\n",is);
    for(int ib = 0; ib < _nb[is]; ++ib) {
      /* Pad each shot in time */
      wp.set_time(stimej[ib]);
      wp.forward(false,_src,tpd);
      fprintf(stderr,"xs=%f stime=%f\n",xsi[ib],stimej[ib]);
      zeropad zp = zeropad(static_cast<int>(xsi[ib]),_srcz);
      /* Blend the shots */
      zp.forward(true,tpd,prc);
    }
    /* Compute background wavefield */
    ss.forward(false,prc,ssc);
    fd.forward(false,ssc,prc);
    /* If user specifies, write out background wavefield */
    if(!_bwfname.empty()) {
      if(is%_wfldj == 0) {
        _wfout.write_all(_bwfname,prc);
        _wfout.write_axis(_bwfname,4,axis(bw+1,0,1.0,"shot"));
        ++bw;
      }
    }
    /* Compute data perturbation */
    memcpy(sht->vals,&d->vals[2*_nt*_nrx*is],sizeof(float)*2*_nt*_nrx);
    /* Back propagate perturbation */
    pvrr.forward(false,sht,ssc);
    fd.adjoint(false,bck,ssc);
    /* If user specifies, write out adjoint wavefield */
    if(!_awfname.empty()) {
      if(is%_wfldj == 0) {
        _awfout.write_all(_awfname,bck);
        _awfout.write_axis(_awfname,4,axis(aw+1,0,1.0,"shot"));
        ++aw;
      }
    }
    ss.adjoint(false,ssc,bck);
    /* Imaging condition */
    d2.adjoint(false,bck,ssc);
    im.adjoint(false,img,bck);
    memcpy(&_imgs->vals[is*_nx*_nz],img->vals,sizeof(float)*_nx*_nz);
    delete bck; delete ssc; delete img;
    delete sht; delete rec; delete tpd;
  }

  /* Stack over all images */
  hypercube_float *timg = new hypercube_float(_bvel->get_axes(),true);
  for(int ii = 0; ii < (int)_nb.size(); ++ii) {
    memcpy(timg->vals,&_imgs->vals[ii*_nx*_nz],sizeof(float)*_nx*_nz);
    m->add(timg);
  }

  return true;
}
