/**
 * Linearized waveform inversion
 * of blended acoustic particle velocity data
 * @author: Joseph Jennings
 * @version: 2017.05.01
 */

#include "sep_params.h"
#include "../main/bbornpv.h"
#include "../main/conj_grad.h"
#include "../main/grad_taper.h"
#include "../main/simpleplot.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc,argv);
  int niter = pars.get_int("niter",10);
  int wfldj = pars.get_int("wfldwj",10);
  /* Source/Receiver axes */
  float srcz = pars.get_float("srcz",50.0);
  float recz = pars.get_float("recz",50.0);
  /* Wavefield sampling */
  float dtu  = pars.get_float("dtu",0.001);
  float fmax = pars.get_float("fmax",50);
  float rho  = pars.get_float("rho",1000.0);
  /* Blending parameters */
  std::vector<int> vminblnd = pars.get_auxvints("minblnd",1,"in");
  std::vector<int> vmaxblnd = pars.get_auxvints("minblnd",1,"in");
  std::vector<float> stimes = pars.get_auxvfloats("stimes",100,"in");
  std::vector<float> xss    = pars.get_auxvfloats("xss",100,"in");
  int minblnd = vminblnd[0]; int maxblnd = vmaxblnd[0];
  /* Taper parameters */
  float z1t = pars.get_float("z1t",500);
  float z2t = pars.get_float("z2t",1500);

  /* Input files */
  sregf datin = sregf("in");
  sregf srcin = sregf("src");
  sregf velin = sregf("bvel");

  /* Outputs file for inversion movies */
  std::string mmov = pars.get_string("mmov","");
  std::string gmov = pars.get_string("gmov","");
  std::string rmov = pars.get_string("rmov","");
  std::string ofno = pars.get_string("ofno","");

  /* Output file for wavefields */
  std::string bwfname = pars.get_string("bwfldsout","");
  std::string awfname = pars.get_string("awfldsout","");

  std::vector<axis>paxes;
  paxes.push_back(velin.get_axis(1));
  paxes.push_back(velin.get_axis(2));
  paxes.push_back(srcin.get_axis(1));

  /* Build hypercubes */
  hypercube_float *vel = new hypercube_float(velin.get_axes(),true);
  hypercube_float *src = new hypercube_float(srcin.get_axes(),true);
  hypercube_float *dat = new hypercube_float(datin.get_axes(),true);
  hypercube_float *mod = new hypercube_float(velin.get_axes(),true);

  /* Read in input files */
  datin.read_all("in",dat);
  srcin.read_all("src",src);
  velin.read_all("bvel",vel);

  /* Construct operator */
  bbornpv *bborn = new bbornpv(paxes,datin.get_axes(),srcz,recz,vel,src,rho,dtu,fmax,minblnd,maxblnd,stimes,xss);

  /* Gradient taper operator */
  grad_taper *tap = new grad_taper(velin.get_axis(1),z1t,z2t);

  conj_grad cg = conj_grad(niter);
  if(!mmov.empty()) {
    cg.set_modout("mmov");
  } if(!gmov.empty()) {
    cg.set_grdout("gmov");
  } if(!rmov.empty()) {
    cg.set_resout("rmov");
  } if(!ofno.empty()) {
    cg.set_ofout("ofno");
  }
  cg.run(bborn,tap,dat,mod);

  /* Write out inverted model */
  sregf modout = sregf(mod->get_axes());
  modout.write_description("out");
  modout.write_all("out",mod);
  int one = 1;
  putch("n3","d",&one);
  putch("n4","d",&one);

  /* Clean up */
  delete vel; delete src; delete dat; delete mod;
}
