/**
 * Linearized waveform inversion
 * with polarization filters
 * @author: Joseph Jennings
 * @version: 2017.05.14
 */

#include "sep_params.h"
#include "../main/indbornpv.h"
#include "../main/mcpefdblnd.h"
#include "../main/grad_taper.h"
#include "../main/cgstep.h"
#include "../main/cghelp.h"
#include "../main/simpleplot.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc,argv);
  int niter = pars.get_int("niter",10);
  /* Source/Receiver axes */
  float srcz = pars.get_float("srcz",50.0);
  float recz = pars.get_float("recz",50.0);
  /* Wavefield sampling */
  float dtu  = pars.get_float("dtu",0.001);
  float fmax = pars.get_float("fmax",50);
  float rho  = pars.get_float("rho",1000.0);
  /* Blending parameters */
  std::vector<int> maxblnd  = pars.get_auxvints("maxblnd",1,"in");
  std::vector<int> minblnd  = pars.get_auxvints("maxblnd",1,"in");
  std::vector<float> xss    = pars.get_auxvfloats("xss",100,"in");
  std::vector<float> stimes = pars.get_auxvfloats("stimes",100,"in");
  /* Taper parameters */
  float z1t = pars.get_float("z1t",500);
  float z2t = pars.get_float("z2t",1500);
  /* PEF parameters */
  int pefniter          = pars.get_int("pefniter",600);
  int sniniter          = pars.get_int("sniniter",300);
  float eps             = pars.get_float("eps",0.01);
  float shtscale        = pars.get_float("scale",10000);
  std::vector<int> j    = pars.get_vints("j",3);
  std::vector<int> nlag = pars.get_vints("nlag",3);
  int pefthrds          = pars.get_int("pefthrds",23);
  /* Check PEF parameters */
  if(nlag.empty()) {
    seperr("Please supply either lag coordinates or the filter box dimensions\n");
  }
  if(nlag.size() != 3) {
    seperr("Please supply three nlag values or a list of lag coordinates\n");
  }
  if(j.empty()) {
    seperr("Please supply filter coarse sampling (j's).\n");
  }
  if(j.size() != 3) {
    seperr("Please supply three j values\n");
  }

  /* Input files */
  sregf datin = sregf("in");
  sregf srcin = sregf("src");
  sregf velin = sregf("bvel");
  sregf copin = sregf("cdat");

  /* Outputs file for inversion movies */
  std::string mmov = pars.get_string("mmov","");
  std::string gmov = pars.get_string("gmov","");
  std::string rmov = pars.get_string("rmov","");
  std::string dmov = pars.get_string("dmov","");
  std::string ofno = pars.get_string("ofno","");

  /* Output file for wavefields */
  std::string bwfname = pars.get_string("bwfldsout","");
  std::string awfname = pars.get_string("awfldsout","");

  int nt  = datin.get_axis(1).n;
  int nrx = datin.get_axis(2).n;
  int nc  = datin.get_axis(3).n;
  int nsx = (int)xss.size();

  std::vector<axis> paxes, iaxes, shaxes;
  paxes.push_back(velin.get_axis(1));
  paxes.push_back(velin.get_axis(2));
  paxes.push_back(srcin.get_axis(1));

  axis saxis = axis(nsx,0.0,1.0);
  iaxes.push_back(datin.get_axis(1));
  iaxes.push_back(datin.get_axis(2));
  iaxes.push_back(datin.get_axis(3));
  iaxes.push_back(saxis);

  /* Build hypercubes */
  hypercube_float *vel  = new hypercube_float(velin.get_axes(),true);
  hypercube_float *src  = new hypercube_float(srcin.get_axes(),true);
  hypercube_float *bdat = new hypercube_float(datin.get_axes(),true);
  hypercube_float *mod  = new hypercube_float(velin.get_axes(),true);
  hypercube_float *idat = new hypercube_float(iaxes,true);
  hypercube_float *ddat = new hypercube_float(iaxes,true);
  hypercube_float *cdat = new hypercube_float(iaxes,true);

  /* Build inversion vectors */
  hypercube_float *res   = new hypercube_float(iaxes,true);
  hypercube_float *drs   = new hypercube_float(iaxes,true);
  hypercube_float *grd   = new hypercube_float(velin.get_axes(),true);
  hypercube_float *grdtp = new hypercube_float(velin.get_axes(),true);

  /* Read in input files */
  datin.read_all("in",bdat);
  srcin.read_all("src",src);
  velin.read_all("bvel",vel);
  copin.read_all("cdat",cdat);

  /* Construct independent modeling operator */
  indbornpv indborn = indbornpv(paxes,iaxes,srcz,recz,vel,src,rho,dtu,fmax,stimes,xss);

  /* Gradient taper operator */
  grad_taper tap = grad_taper(velin.get_axis(1),z1t,z2t);

  /* Polarization filter */
  //XXX: Don't forget the option to pad instead of using default
  mcpefdblnd pf = mcpefdblnd(pefniter,sniniter,eps,j,nlag,shtscale,minblnd[0],maxblnd[0],nsx,datin.get_axes(),pefthrds);

  /* Initialize */
  mod->zero(); idat->zero(); res->zero();

  /* Build CG stepper */
  cgstep step = cgstep();
  step.alloc_step(mod,(norm_vector*)idat);

  /* CG output utility */
  cghelp dgn = cghelp(mmov,gmov,ofno,dmov);

  /* Conjugate gradient inversion */
  for(int iter = 0; iter < niter; ++iter) {
    /* First compute the objective function */
    fprintf(stderr,"Computing residual...\n");
    indborn.forward(false, mod, idat);
    /* First iteration operates on blended data */
    if(iter == 0) {
      memcpy(ddat->vals,cdat->vals,sizeof(float)*nt*nrx*nc*nsx);
    } else {
      fprintf(stderr,"Apply polarization filter...\n");
      pf.apply(bdat,idat,ddat);
      fprintf(stderr,"ddat max=%f\n",ddat->my_max());
    }
    res->scale_add(1.0,idat,-1.0,ddat);
    double f0 = (0.5)*(res->dot(res));
    /* Compute the gradient*/
    fprintf(stderr,"Computing gradient...\n");
    indborn.adjoint(false, grd, res);
    /* Taper the gradient */
    tap.forward(false,grd,grdtp);
    fprintf(stderr,"grdtp max=%f\n",grdtp->my_max());
    /* Compute data space gradient */
    fprintf(stderr,"Computing data-space gradient...\n");
    indborn.forward(false, grdtp, drs);
    fprintf(stderr,"drs max=%f\n",drs->my_max());
    /* Compute the step length and update*/
    if(!step.step_it(mod,grdtp,res,drs)) break;
    /* Reevaluate the objective function and output */
    double f1 = (0.5)*res->dot(res);
    if(f1 >= f0) {
      fprintf(stderr, "Objective function did not reduce, will terminate solver.\n");
      break;
    }
    dgn.output(res,ddat,grdtp,mod,f1);
    fprintf(stderr,"iter=\t%d\t obj=\t%g\n",iter+1,f1);
  }

  /* Write out inverted model */
  sregf imodout = sregf(mod->get_axes());
  imodout.write_description("out");
  imodout.write_all("out",mod);
  int one = 1;
  putch("n3","d",&one);
  putch("n4","d",&one);

}
