/**
 * Blended modeling of particle velocity
 * acoustic data. Blends wavelets and then
 * models the blended source function.
 * @author: Joseph Jennings
 * @version: 2017.05.14
 */

#include <random>
#include <omp.h>
#include "sep_params.h"
#include "sourcescale.h"
#include "zeropad.h"
#include "laplacian10th.h"
#include "laplacian10thispc.h"
#include "afd2dfst.h"
#include "../main/pvrrec.h"
#include "../main/waveletpad.h"
#include "../main/blendnc2d.h"
#include "../main/simpleplot.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc, argv);
  int verb = pars.get_int("verb",0);
  int nthd = pars.get_int("nthrds",2);
  /* Shot axes */
  int nsx    = pars.get_int  ("nsx",  1);
  float osx  = pars.get_float("osx",0.0);
  float dsx  = pars.get_float("dsx",1.0);
  float srcz = pars.get_float("srcz",50.0);
  std::vector<float> sxs = pars.get_vfloats("sxs",10);
  std::vector<float> szs = pars.get_vfloats("szs",10);
  if(!sxs.empty() && !szs.empty()) nsx = sxs.size();
  /* Receiver axes */
  int nrx    = pars.get_int  ("nrx",  1);
  float orx  = pars.get_float("orx",0.0);
  float drx  = pars.get_float("drx",1.0);
  float recz = pars.get_float("recz",50.0);
  /* Wavefield sampling */
  float dtu = pars.get_float("dtu",0.001);
  float fmax = pars.get_float("fmax",50);
  /* Density at receivers */
  float rho = pars.get_float("rho",1000.0);
  /* Blending parameters */
  int minblnd  = pars.get_int("minblnd",2);
  int maxblnd  = pars.get_int("maxblnd",3);
  float maxdly = pars.get_float("maxdly",0.5);
  float mindly = pars.get_float("mindly",0.0);
  int seed     = pars.get_int("seed",0);
  int space    = pars.get_int("space",7);
  std::string enctag = pars.get_string("encsrcs","");
  std::vector<float> stimei = pars.get_vfloats("stimes",100);
  int wfldwj = pars.get_int("wfldwj",1);

  /* Output file for wavefields */
  std::string wfname = pars.get_string("wfldsout","");

  /* Build random shot times if none provided*/
  if(stimei.empty()) {
    std::random_device rd; std::mt19937 rng;
    if(!seed) {
      rng.seed(rd());
    } else {
      rng.seed(seed);
    }
    std::uniform_real_distribution<float> unif(maxdly,mindly);
    /* First time is always zero */
    stimei.push_back(0.0);
    for(int is = 1; is < nsx; ++is) {
      auto val = unif(rng);
      stimei.push_back(val);
    }
  } else {
    if((int)stimei.size() != nsx) {
      seperr("Number of shot times %d must equal number of shots %d\n",(int)stimei.size(),nsx);
    }
  }

  /* Build blending array */
  blendnc2d *gam = new blendnc2d(nsx,minblnd,maxblnd);
  std::vector<int> nb = gam->get_nb();

  /* Input files */
  sregf srcin = sregf("in");
  sregf velin = sregf("vel");
  sregf wfout;

  int   nz = velin.get_axis(1).n;
  int   nx = velin.get_axis(2).n;
  int   nt = srcin.get_axis(1).n;

  std::vector<axis> daxes,raxes;
  daxes.push_back(srcin.get_axis(1)); raxes.push_back(srcin.get_axis(1));
  daxes.push_back(axis(nrx,orx,drx)); raxes.push_back(axis(nrx,orx,drx));
  daxes.push_back(axis(2,0.0,1.0));   raxes.push_back(axis(2,0.0,1.0));
  daxes.push_back(axis((int)nb.size(),0.0,1.0));

  std::vector<axis> paxes;
  /* Wavefield axes */
  paxes.push_back(velin.get_axis(1));
  paxes.push_back(velin.get_axis(2));
  paxes.push_back(srcin.get_axis(1));

  /* Build hypercubes */
  hypercube_float *src = new hypercube_float(srcin.get_axes(),true);
  hypercube_float *vel = new hypercube_float(velin.get_axes(),true);
  hypercube_float *dat = new hypercube_float(daxes,true);

  /* Read in input files */
  std::vector<int> nwind,fwind;
  nwind.push_back(nz); nwind.push_back(nx);
  nwind.push_back(nt); nwind.push_back(1);
  fwind.push_back(0); fwind.push_back(0);
  fwind.push_back(0); fwind.push_back(0);
  srcin.read_all("in",src);
  velin.read_all("vel",vel);

  if(!wfname.empty()) {
    wfout = sregf(paxes);
    wfout.write_description(wfname);
  }

  /* Construct operators */
  sourcescale ss          = sourcescale(dtu,vel);

  /* Create shot positions */
  std::vector<std::vector<float>> xss;
  std::vector<std::vector<float>> stimes;
  int j = 0; int k = space;
  for(int is = 0; is < (int)nb.size(); ++is) {
    std::vector<float> xsi;
    std::vector<float> stimej;
    if(is != 0 && is%space == 0)  j += k;
    for(int ib = 0; ib < nb[is]; ++ib) {
      xsi.push_back(osx + dsx*j);
      stimej.push_back(stimei[j]);
      if(ib == 0) {
        if( j + k >= nsx)  k = 2; //FIXME: A hack for now
        j += k;
      }
    }
    j = j - k + 1;
    xss.push_back(xsi);
    stimes.push_back(stimej);
  }

  /* Loop over shot */
  int iw = 0;
  omp_set_num_threads(nthd);
#pragma omp parallel for default(shared)
  for(int is = 0; is < (int)nb.size(); ++is) {
    waveletpad wp           = waveletpad(src->get_axis(1));
    laplacian10thispc *lp   = new laplacian10thispc(paxes);
    afd2dfst fd             = afd2dfst(paxes,vel,lp,dtu,fmax);
    pvrrec pvrr             = pvrrec(paxes,axis(nrx,orx,drx),recz,rho);
    hypercube_float *tpd = new hypercube_float(srcin.get_axes(),true);
    hypercube_float *prc = new hypercube_float(paxes,true);
    hypercube_float *ssc = new hypercube_float(paxes,true);
    hypercube_float *rec = new hypercube_float(raxes,true);
    if(verb) fprintf(stderr,"shotnum=%d\n",is);
    /* Build blended source */
    prc->zero();
    std::vector<float> xsi = xss[is];
    std::vector<float> stimej = stimes[is];
    for(int ib = 0; ib < nb[is]; ++ib) {
      fprintf(stderr,"xs=%f stime=%f\n",xsi[ib],stimej[ib]);
      /* Pad each shot in time */
      wp.set_time(stimej[ib]);
      wp.forward(false,src,tpd);
      zeropad zp = zeropad(static_cast<int>(xsi[ib]),static_cast<int>(srcz));
      /* Blend the shots */
      zp.forward(true,tpd,prc);
    }
    ss.forward(false,prc,ssc);
    fd.forward(false,ssc,prc);
    /* If user specifies, write out wavefield */
    if(!wfname.empty()) {
      if(is%wfldwj == 0) {
        wfout.write_all(wfname,prc);
        wfout.write_axis(wfname,4,axis(iw+1,0,1.0,"shot"));
        ++iw;
      }
    }
    /* Multicomponent receiver restriction */
    pvrr.adjoint(false,rec,prc);
    memcpy(&dat->vals[2*is*nt*nrx],rec->vals,sizeof(float)*2*nt*nrx);
    delete prc; delete ssc; delete rec; delete tpd;
  }

  /* Write out data */
  sregf datout = sregf(dat->get_axes());
  datout.write_description("out");
  datout.write_all("out",dat);

  /* Write out shot times and positions */
  printf("stimes=");
  for(int is = 0; is < (int)nb.size(); ++is) {
    if(nsx%2 != 0 && is == (int)nb.size() - 1) printf("%f",stimes[is][0]);
    else printf("%f,%f,",stimes[is][0],stimes[is][1]);
  }
  printf("\n");
  printf("xss=");
  for(int is = 0; is < (int)nb.size(); ++is) {
    if(nsx%2 != 0 && is == (int)nb.size() - 1) printf("%f",xss[is][0]);
    else printf("%f,%f,",xss[is][0],xss[is][1]);
  }
  printf("\n");
  printf("maxblnd=%d\n",maxblnd);
  printf("minblnd=%d\n",minblnd);

  /* Clean up */
  delete vel; delete dat;
}
