/**
 * Test the adjoint of the
 * 2D acoustic, constant-density
 * finite difference propagator
 * @author: Joseph Jennings
 * @version: 2016.12.7
 */

#include <limits.h>
#include "gtest/gtest.h"
#include "../main/acfdprop2d.h"
#include "../main/zeropadimg.h"
#include "../main/zeropadwfld.h"
#include "../main/laplacian4th.h"
#include "../main/laplacian10th.h"
#include "../main/simpleplot.h"

class acfd2dproptest: public ::testing::Test {
  public:
    void zeroedge(hypercube_float *in, int shift);
  protected:
    int nx, nz, ntu;
    int p; int shift;
    float dx, dz, dtu;
    axis ax, az, axp, azp, atu;
    std::vector<axis> axes, paxes, uaxes, puaxes;
    hypercube_float *rm, *rd;
    virtual void SetUp(){
      /* Geometry parameters */
      nx = 200; nz = 200; ntu = 100;
      dx = 10.0; dz = 10.0; dtu = 0.001;
      /* Build axes */
      ax  = axis(nx, 0.0, dx);
      az  = axis(nz, 0.0, dz);
      atu = axis(ntu , 0.0, dtu);
      axes.push_back(az); axes.push_back(ax);
      uaxes.push_back(az); uaxes.push_back(ax);
      uaxes.push_back(atu);
      rm = new hypercube_float(uaxes,true);
      rd = new hypercube_float(uaxes,true);
      rm->random(); rd->random();
    }

    virtual void TearDown() {
      delete rm; delete rd;
    }
};

TEST_F(acfd2dproptest,dptestnb4){
  /* Padded axes */
  p = 4; shift = p/2;
  axp = axis(nx+p, 0.0, dx);
  azp = axis(nz+p, 0.0, dz);
  paxes.push_back(azp); paxes.push_back(axp);
  /* Wavefield axes */
  puaxes.push_back(azp); puaxes.push_back(axp);
  puaxes.push_back(atu);
  /* Padded velocity */
  hypercube_float *velp = new hypercube_float(paxes, true);
  velp->set_val(3000.0); zeroedge(velp,shift);
  zeropadwfld zpw = zeropadwfld(shift,shift,0.0);
  /* laplacian operator */
  laplacian4th *lap = new laplacian4th(paxes);
  /* Model and data hypercubes */
  hypercube_float *m  = new hypercube_float(puaxes,true);
  hypercube_float *ms = new hypercube_float(puaxes,true);
  hypercube_float *d  = new hypercube_float(puaxes,true);
  hypercube_float *ds = new hypercube_float(puaxes,true);
  zpw.forward(false,rm,m); zpw.forward(false,rd,d);
  /* Wave propagation operator */
  acfdprop2d prp = acfdprop2d(puaxes,velp,lap,60.0);
  /* Add == false */
  prp.forward(false,m,ds);
  float dotd = d->dot(ds);
  prp.adjoint(false,ms,d);
  float dotm = m->dot(ms);
  ASSERT_NEAR(dotm,dotd,0.01);
  /* Clean up */
  delete velp;
  delete m; delete ms;
  delete d; delete ds;
}

TEST_F(acfd2dproptest,dptestnb10){
  /* Padded axes */
  p = 10; shift = p/2;
  axp = axis(nx+p, 0.0, dx);
  azp = axis(nz+p, 0.0, dz);
  paxes.push_back(azp); paxes.push_back(axp);
  /* Wavefield axes */
  puaxes.push_back(azp); puaxes.push_back(axp);
  puaxes.push_back(atu);
  /* Velocity and padded velocity */
  hypercube_float *velp = new hypercube_float(paxes, true);
  velp->set_val(3000.0); zeroedge(velp,shift);
  zeropadwfld zpw = zeropadwfld(shift,shift,0.0);
  /* laplacian operator */
  laplacian10th *lap = new laplacian10th(paxes);
  /* Model and data hypercubes */
  hypercube_float *m  = new hypercube_float(puaxes,true);
  hypercube_float *ms = new hypercube_float(puaxes,true);
  hypercube_float *d  = new hypercube_float(puaxes,true);
  hypercube_float *ds = new hypercube_float(puaxes,true);
  zpw.forward(false,rm,m); zpw.forward(false,rd,d);
  /* Wave propagation operator */
  acfdprop2d prp = acfdprop2d(puaxes,velp,lap,60.0);
  /* Add == false */
  prp.forward(false,m,ds);
  float dotd = d->dot(ds);
  prp.adjoint(false,ms,d);
  float dotm = m->dot(ms);
  ASSERT_NEAR(dotm,dotd,0.000001);
  /* Clean up */
  delete velp;
  delete m; delete ms;
  delete d; delete ds;
}

TEST_F(acfd2dproptest,dptestb4){
  /* Padded axes */
  p = 4; shift = p/2;
  axp = axis(nx+p, 0.0, dx);
  azp = axis(nz+p, 0.0, dz);
  /* Velocity axes */
  axes.push_back(az); axes.push_back(ax);
  paxes.push_back(azp); paxes.push_back(axp);
  /* Wavefield axes */
  puaxes.push_back(azp); puaxes.push_back(axp);
  puaxes.push_back(atu);
  /* Absorbing layer */
  std::vector<int> nbs;
  int bz1=20; int bz2=20; int bx1=20; int bx2=20;
  nbs.push_back(bz1); nbs.push_back(bx1);
  nbs.push_back(bz2); nbs.push_back(bx2);
  /* Velocity and padded velocity */
  hypercube_float *velp = new hypercube_float(paxes,true);
  velp->set_val(3000.0); zeroedge(velp,shift);
  zeropadwfld zpw = zeropadwfld(shift,shift,0.0);
  /* Weights and padded weights */
  hypercube_float *wtp   = new hypercube_float(paxes, true);
  hypercube_float *vwtp = new hypercube_float(paxes, true);
  wtp->set_val(0.01); vwtp->set_val(0.1);
  zeroedge(wtp,shift); zeroedge(vwtp,shift);
  /* laplacian operator */
  laplacian4th *lap = new laplacian4th(paxes);
  /* Model and data hypercubes */
  hypercube_float *m  = new hypercube_float(puaxes,true);
  hypercube_float *ms = new hypercube_float(puaxes,true);
  hypercube_float *d  = new hypercube_float(puaxes,true);
  hypercube_float *ds = new hypercube_float(puaxes,true);
  zpw.forward(false,rm,m); zpw.forward(false,rd,d);
  /* Wave propagation operator */
  acfdprop2d prp = acfdprop2d(puaxes,velp,wtp,vwtp,nbs,lap,60.0);
  /* Add == false */
  prp.forward(false,m,ds);
  float dotd = d->dot(ds);
  prp.adjoint(false,ms,d);
  float dotm = m->dot(ms);
  ASSERT_NEAR(dotm,dotd,0.001);
  /* Clean up */
  delete wtp; delete vwtp;
  delete velp;
  delete m; delete ms;
  delete d; delete ds;
}

TEST_F(acfd2dproptest,dptestb10){
  /* Padded axes */
  p = 10; shift = p/2;
  axp = axis(nx+p, 0.0, dx);
  azp = axis(nz+p, 0.0, dz);
  /* Velocity axes */
  axes.push_back(az); axes.push_back(ax);
  paxes.push_back(azp); paxes.push_back(axp);
  /* Wavefield axes */
  puaxes.push_back(azp); puaxes.push_back(axp);
  puaxes.push_back(atu);
  /* Absorbing layer */
  std::vector<int> nbs;
  int bz1=20; int bz2=20; int bx1=20; int bx2=20;
  nbs.push_back(bz1); nbs.push_back(bx1);
  nbs.push_back(bz2); nbs.push_back(bx2);
  /* Velocity and padded velocity */
  hypercube_float *velp = new hypercube_float(paxes,true);
  velp->set_val(3000.0); zeroedge(velp,shift);
  zeropadwfld zpw = zeropadwfld(shift,shift,0.0);
  /* Weights and padded weights */
  hypercube_float *wtp  = new hypercube_float(paxes, true);
  hypercube_float *vwtp = new hypercube_float(paxes, true);
  /* laplacian operator */
  laplacian10th *lap = new laplacian10th(paxes);
  /* Model and data hypercubes */
  hypercube_float *m  = new hypercube_float(puaxes,true);
  hypercube_float *ms = new hypercube_float(puaxes,true);
  hypercube_float  *d  = new hypercube_float(puaxes,true);
  hypercube_float  *ds = new hypercube_float(puaxes,true);
  zpw.forward(false,rm,m); zpw.forward(false,rd,d);
  /* Wave propagation operator */
  acfdprop2d prp = acfdprop2d(puaxes,velp,wtp,vwtp,nbs,lap,60.0);
  /* Add == false */
  prp.forward(false,m,ds);
  float dotd = d->dot(ds);
  prp.adjoint(false,ms,d);
  float dotm = m->dot(ms);
  ASSERT_NEAR(dotm,dotd,0.000001);
  /* Clean up */
  delete wtp; delete vwtp;
  delete velp;
  delete m; delete ms;
  delete d; delete ds;
}

void acfd2dproptest::zeroedge(hypercube_float *in, int shift) {

  int n1 = in->get_axis(1).n; int n2 = in->get_axis(2).n;
  /* Top edge */
    for(int i2 = 0; i2 < n2; ++i2) {
      for(int i1 = 0; i1 < shift; ++i1) {
        in->vals[i2*n1 + i1] = 0.0;
    }
  }

  /* Left edge */
  for(int i2 = 0; i2 < shift; ++i2) {
    for(int i1 = 0; i1 < n1; ++i1) {
      in->vals[i2*n1 + i1] = 0.0;
    }
  }

  /* Right edge */
  for(int i2 = n2-shift; i2 < n2; ++i2) {
    for(int i1 = 0; i1 < n1; ++i1) {
      in->vals[i2*n1 + i1] = 0.0;
    }
  }

  /* Bottom edge */
  for(int i2 = 0; i2 < n2; ++i2) {
    for(int i1 = n1-shift; i1 < n1; ++i1) {
      in->vals[i2*n1 + i1] = 0.0;
    }
  }
}
