/**
 * Test the adjoint of the fast 2D
 * acoustic, constant density
 * wave-equation solver
 * @author: Joseph Jennings
 * @version: 2017.01.7
 */

#include <limits.h>
#include "gtest/gtest.h"
#include "../main/afd2dfst.h"
#include "../main/zeropadimg.h"
#include "../main/zeropadwfld.h"
#include "../main/laplacian4th.h"
#include "../main/laplacian10th.h"
#include "../main/laplacian10thispc.h"
#include "../main/simpleplot.h"

class afd2dfstproptest: public ::testing::Test {
  public:
    void zeroedge(hypercube_float *in, int shift);
  protected:
    int nx, nz, ntu, nts;
    int p; int shift;
    float dx, dz, dtu, dts;
    axis ax, az, axp, azp, atu, ats;
    std::vector<axis> axes, paxes, uaxes, saxes, puaxes;
    hypercube_float *rm, *rd;
    virtual void SetUp(){
      /* Geometry parameters */
      nx = 200; nz = 200; ntu = 100; nts = 50;
      dx = 10.0; dz = 10.0; dtu = 0.001; dts = 0.002;
      /* Build axes */
      ax  = axis(nx, 0.0, dx);
      az  = axis(nz, 0.0, dz);
      atu = axis(nts , 0.0, dtu);
      ats = axis(nts, 0.0, dts);
      axes.push_back(az); axes.push_back(ax);
      uaxes.push_back(az); uaxes.push_back(ax);
      uaxes.push_back(atu);
      saxes.push_back(az); saxes.push_back(ax);
      saxes.push_back(ats);
      rm = new hypercube_float(saxes,true);
      rd = new hypercube_float(saxes,true);
      rm->random(); rd->random();
    }

    virtual void TearDown() {
      //delete rm; delete rd;
    }
};

TEST_F(afd2dfstproptest,dptest4){
  /* Padded axes */
  p = 4; shift = p/2;
  axp = axis(nx+p, 0.0, dx);
  azp = axis(nz+p, 0.0, dz);
  paxes.push_back(azp); paxes.push_back(axp);
  /* Wavefield axes */
  puaxes.push_back(azp); puaxes.push_back(axp);
  puaxes.push_back(ats);
  /* Padded velocity */
  hypercube_float *velp = new hypercube_float(paxes, true);
  velp->set_val(3000.0); zeroedge(velp,shift);
  zeropadwfld zpw = zeropadwfld(shift,shift,0.0);
  /* laplacian operator */
  laplacian4th *lap = new laplacian4th(paxes);
  /* Model and data hypercubes */
  hypercube_float *m  = new hypercube_float(puaxes,true);
  hypercube_float *ms = new hypercube_float(puaxes,true);
  hypercube_float *d  = new hypercube_float(puaxes,true);
  hypercube_float *ds = new hypercube_float(puaxes,true);
  zpw.forward(false,rm,m); zpw.forward(false,rd,d);
  /* Wave propagation operator */
  afd2dfst *prp = new afd2dfst(puaxes,velp,lap,dtu,60.0);
  /* Add == false */
  prp->forward(false,m,ds);
  float dotd = d->dot(ds);
  prp->adjoint(false,ms,d);
  float dotm = m->dot(ms);
  ASSERT_NEAR(dotm,dotd,0.0001);
  /* Clean up */
  delete velp;
  delete m; delete ms;
  delete d; delete ds;
}

TEST_F(afd2dfstproptest,dptest10){
  /* Padded axes */
  p = 10; shift = p/2;
  axp = axis(nx+p, 0.0, dx);
  azp = axis(nz+p, 0.0, dz);
  paxes.push_back(azp); paxes.push_back(axp);
  /* Wavefield axes */
  puaxes.push_back(azp); puaxes.push_back(axp);
  puaxes.push_back(ats);
  /* Velocity and padded velocity */
  hypercube_float *velp = new hypercube_float(paxes, true);
  velp->set_val(3000.0); zeroedge(velp,shift);
  zeropadwfld zpw = zeropadwfld(shift,shift,0.0);
  /* laplacian operator */
  laplacian10th *lap = new laplacian10th(paxes);
  /* Model and data hypercubes */
  hypercube_float *m  = new hypercube_float(puaxes,true);
  hypercube_float *ms = new hypercube_float(puaxes,true);
  hypercube_float *d  = new hypercube_float(puaxes,true);
  hypercube_float *ds = new hypercube_float(puaxes,true);
  zpw.forward(false,rm,m); zpw.forward(false,rd,d);
  /* Wave propagation operator */
  afd2dfst *prp =  new afd2dfst(puaxes,velp,lap,dtu,60.0);
  /* Add == false */
  prp->forward(false,m,ds);
  float dotd = d->dot(ds);
  prp->adjoint(false,ms,d);
  float dotm = m->dot(ms);
  //TODO: check relative error (dotd-dotm)/dotm
  //TODO: Also check the magnitude of the vectors
  ASSERT_NEAR(dotm,dotd,0.001);
  /* Clean up */
  delete velp;
  delete m; delete ms;
  delete d; delete ds;
}
TEST_F(afd2dfstproptest,dptestispc){
  /* Padded axes */
  p = 10; shift = p/2;
  axp = axis(nx+p, 0.0, dx);
  azp = axis(nz+p, 0.0, dz);
  paxes.push_back(azp); paxes.push_back(axp);
  /* Wavefield axes */
  puaxes.push_back(azp); puaxes.push_back(axp);
  puaxes.push_back(ats);
  /* Velocity and padded velocity */
  hypercube_float *velp = new hypercube_float(paxes, true);
  velp->set_val(3000.0); zeroedge(velp,shift);
  zeropadwfld zpw = zeropadwfld(shift,shift,0.0);
  /* laplacian operator */
  laplacian10thispc *lap = new laplacian10thispc(paxes);
  /* Model and data hypercubes */
  hypercube_float *m  = new hypercube_float(puaxes,true);
  hypercube_float *ms = new hypercube_float(puaxes,true);
  hypercube_float *d  = new hypercube_float(puaxes,true);
  hypercube_float *ds = new hypercube_float(puaxes,true);
  zpw.forward(false,rm,m); zpw.forward(false,rd,d);
  /* Wave propagation operator */
  afd2dfst *prp =  new afd2dfst(puaxes,velp,lap,dtu,60.0);
  /* Add == false */
  prp->forward(false,m,ds);
  float dotd = d->dot(ds);
  prp->adjoint(false,ms,d);
  float dotm = m->dot(ms);
  //TODO: check relative error (dotd-dotm)/dotm
  //TODO: Also check the magnitude of the vectors
  ASSERT_NEAR(dotm,dotd,0.001);
  /* Clean up */
  delete velp;
  delete m; delete ms;
  delete d; delete ds;
}

void afd2dfstproptest::zeroedge(hypercube_float *in, int shift) {

  int n1 = in->get_axis(1).n; int n2 = in->get_axis(2).n;
  /* Top edge */
    for(int i2 = 0; i2 < n2; ++i2) {
      for(int i1 = 0; i1 < shift; ++i1) {
        in->vals[i2*n1 + i1] = 0.0;
    }
  }

  /* Left edge */
  for(int i2 = 0; i2 < shift; ++i2) {
    for(int i1 = 0; i1 < n1; ++i1) {
      in->vals[i2*n1 + i1] = 0.0;
    }
  }

  /* Right edge */
  for(int i2 = n2-shift; i2 < n2; ++i2) {
    for(int i1 = 0; i1 < n1; ++i1) {
      in->vals[i2*n1 + i1] = 0.0;
    }
  }

  /* Bottom edge */
  for(int i2 = 0; i2 < n2; ++i2) {
    for(int i1 = n1-shift; i1 < n1; ++i1) {
      in->vals[i2*n1 + i1] = 0.0;
    }
  }
}
