/**
 * Tests the dot product of the
 * finite-difference absorbing
 * boundary coefficient operator
 * @author: Joseph Jennings
 * @version: 2016.12.27
 */

#include <limits.h>
#include "gtest/gtest.h"
#include "../main/fdcoefb.h"
#include "../main/zeropadimg.h"
#include "../main/laplacian4th.h"
#include "../main/laplacian10th.h"
#include "../main/simpleplot.h"
#include "hypercube_float.h"

class fdcoefbtest: public ::testing::Test {
  protected:
    virtual void SetUp(){
    }

    virtual void TearDown() {
    }
};

/* Fourth order laplacian test */
TEST_F(fdcoefbtest,dptest4){
  int nx = 100; int nz = 100;
  int p = 4;
  float dt2 = (0.001*0.001);
  /* Build axes */
  axis ax  = axis(nx  , 0.0, 1.0);
  axis az  = axis(nz  , 0.0, 1.0);
  axis axp = axis(nx+p, 0.0, 1.0);
  axis azp = axis(nz+p, 0.0, 1.0);
  std::vector<axis> axes, paxes;
  axes.push_back(az); axes.push_back(ax);
  paxes.push_back(azp); paxes.push_back(axp);
  /* Absorbing layer */
  std::vector<int> nbs;
  int bz1=20; int bz2=20; int bx1=20; int bx2=20;
  nbs.push_back(bz1); nbs.push_back(bx1);
  nbs.push_back(bz2); nbs.push_back(bx2);
  /* Model and data for dot product */
  hypercube_float *rm  = new hypercube_float(axes , true);
  hypercube_float *rd  = new hypercube_float(axes , true);
  hypercube_float *m   = new hypercube_float(paxes, true);
  hypercube_float *ms  = new hypercube_float(paxes, true);
  hypercube_float *d   = new hypercube_float(paxes, true);
  hypercube_float *ds  = new hypercube_float(paxes, true);
  /* Auxiliary hypercubes */
  hypercube_float *v   = new hypercube_float(paxes, true);
  hypercube_float *wt  = new hypercube_float(paxes, true);
  hypercube_float *vwt = new hypercube_float(paxes, true);
  /* Initialize hypercubes */
  rm->random(); rd->random();
  v->set_val(3000); v->mult(v); v->scale(dt2);
  wt->set_val(0.01); vwt->set_val(0.1);
  zeropadimg pad = zeropadimg(2,2,0.0);
  laplacian4th *lap = new laplacian4th(paxes);
  fdcoefb coef = fdcoefb(paxes,lap,v,wt,vwt,nbs);
  /* Pad the wavefield (for the laplacian) */
  pad.forward(false,rd,d); pad.forward(false,rm,m);
  /* Add == false */
  coef.forward(false,m,ds);
  float dotd = d->dot(ds);
  coef.adjoint(false,ms,d);
  float dotm = m->dot(ms);
  /* Decreased precision for running all tests */
  EXPECT_NEAR(dotd, dotm, 0.00001);
  delete m; delete ms;
  delete d; delete ds;
  delete rm; delete rd;
  delete v;
}

/* Tenth order laplacian test */
TEST_F(fdcoefbtest,dptest10){
  int nx = 100; int nz = 100;
  int p = 10;
  float dt2 = (0.001*0.001);
  /* Build axes */
  axis ax  = axis(nx  , 0.0, 1.0);
  axis az  = axis(nz  , 0.0, 1.0);
  axis axp = axis(nx+p, 0.0, 1.0);
  axis azp = axis(nz+p, 0.0, 1.0);
  std::vector<axis> axes, paxes;
  axes.push_back(az); axes.push_back(ax);
  paxes.push_back(azp); paxes.push_back(axp);
  /* Absorbing layer */
  std::vector<int> nbs;
  int bz1=20; int bz2=20; int bx1=20; int bx2=20;
  nbs.push_back(bz1); nbs.push_back(bx1);
  nbs.push_back(bz2); nbs.push_back(bx2);
  /* Build hypercubes */
  hypercube_float *rm  = new hypercube_float(axes , true);
  hypercube_float *rd  = new hypercube_float(axes , true);
  hypercube_float *m = new hypercube_float(paxes, true);
  hypercube_float *ms = new hypercube_float(paxes, true);
  hypercube_float *d = new hypercube_float(paxes, true);
  hypercube_float *ds = new hypercube_float(paxes, true);
  /* Auxiliary hypercubes */
  hypercube_float *wt  = new hypercube_float(paxes, true);
  hypercube_float *vwt = new hypercube_float(paxes, true);
  hypercube_float *v  = new hypercube_float(paxes, true);
  /* Initialize hypercubes */
  rm->random(); rd->random();
  v->set_val(3000); v->mult(v); v->scale(dt2);
  wt->set_val(0.01); vwt->set_val(0.1);
  zeropadimg pad = zeropadimg(5,5,0.0);
  laplacian10th *lap = new laplacian10th(paxes);
  fdcoefb coef = fdcoefb(paxes,lap,v,wt,vwt,nbs);
  /* Pad the wavefield (for the laplacian) */
  pad.forward(false,rd,d); pad.forward(false,rm,m);
  /* Add == false */
  coef.forward(false,m,ds);
  float dotd = d->dot(ds);
  coef.adjoint(false,ms,d);
  float dotm = m->dot(ms);
  /* Decreased precision for running all tests */
  EXPECT_NEAR(dotd, dotm, 0.00001);
  delete m; delete ms;
  delete d; delete ds;
  delete rm; delete rd;
  delete v;
}
